from collections import UserDict, UserList
from functools import reduce
from typing import Any, Optional, Tuple
from enum import Enum
from collections import defaultdict

from dicfg.addons.addon import (
    CONFIG_ADDONS,
    Addon,
    TemplateAddon,
    UpdaterAddon,
    ValidatorAddon,
    process_addons,
    select_addon,
)
from dicfg.addons.validators import ValidationError


class Affix(Enum):
    """Affixes for the update function"""

    PRE = "pre"
    POST = "post"


class ConfigValue:
    """Wraps a value into a ConfigValue

    Args:
        data (Any): value of the config
        merger (Callable, optional): Callable to merge the config value. Defaults to None.
    """

    def __init__(
        self,
        data: Any,
        updater: tuple[UpdaterAddon] = None,
        validator: tuple[ValidatorAddon] = None,
    ):
        self.updater = updater or (None,)
        self.validator = validator or (None,)
        self.data = self._init(data)

    def _init(self, data):
        return data

    def modify(self, b: "ConfigValue") -> "ConfigValue":
        """Merges config b with it self

        Args:
            b (ConfigValue): another config

        Returns:
            ConfigValue: self
        """

        for idx, updater in enumerate(self.updater):
            if updater is None and b.updater[idx] is None:
                self.data = update(self, b)
            elif b.updater[idx] is not None:
                self.data = b.updater[idx].update(self, b)
            else:
                self.data = updater.update(self, b)
        return self

    def validate(self):
        """Validate the config"""
        for validator in self.validator:
            if validator is not None:
                if error := validator.validate(self.data):
                    yield error

    def cast(self):
        """Cast wrapped value to builtin python value"""
        return self.data


class ConfigDict(ConfigValue, UserDict[str, ConfigValue]):
    """Wraps a value into a ConfigDict

    Args:
        data (dict): value of the config

    """

    def _init(self, data: dict):
        for key in list(data):
            config_kwargs: dict[str, list[Addon]] = defaultdict(list)
            _key, addons = process_addons(key)
            value = data.pop(key)
            for addon, name in addons:
                config_kwargs[addon].append(select_addon(addon, name))

            templates = config_kwargs.pop(CONFIG_ADDONS.TEMPLATE.value, None)
            data[_key] = _config_factory(value, **config_kwargs)
            if templates is not None:
                data[_key] = self._apply_templates(data[_key], templates)

        return data

    def _apply_templates(
        self, config_value: ConfigValue, templates: list[TemplateAddon]
    ):
        """Apply templates to the given config value."""
        for template in templates:
            template_data = _config_factory(template.data())
            if not isinstance(config_value, type(template_data)):
                config_value = template_data
            else:
                config_value = template_data.modify(config_value)
        return config_value

    def validate(self):
        yield from super().validate()
        for key, value in self.data.items():
            for err in value.validate():
                yield ValidationError(f"{key}:{err.message}")

    def cast(self):
        """Cast wrapped value to builtin python value"""
        return {key: value.cast() for key, value in self.data.items()}


class ConfigList(ConfigValue, UserList):
    """Wraps a value into a ConfigList

    Args:
        data (list): value of the config

    """

    def _init(self, data: list):
        for idx, value in enumerate(data):
            data[idx] = _config_factory(value)
        return data

    def validate(self):
        yield from super().validate()
        for value in self.data:
            yield from value.validate()

    def cast(self):
        """Cast wrapped value to builtin python value"""
        return [value.cast() for value in self.data]


def _config_factory(c, updater=None, validator=None) -> ConfigValue:
    if isinstance(c, ConfigValue):
        return c
    config_types = {dict: ConfigDict, list: ConfigList}
    return config_types.get(type(c), ConfigValue)(
        c, updater=updater, validator=validator
    )


def _insert(dictionary, prev_key, k, v):
    new_dict = {}
    for _k, _v in dictionary.items():
        new_dict[_k] = _v
        if prev_key == _k:
            new_dict[k] = v
    return new_dict


def _modify(a: ConfigValue, b: ConfigValue):
    return a.modify(b)


def update(a: ConfigValue, b: ConfigValue, affix: Optional[Affix] = None) -> Any:
    if not isinstance(b, ConfigDict):
        return b.data

    prev_key = None
    for k, v in b.items():
        if k in a:
            if type(b[k]) != type(a[k]):  # noqa: E721
                a[k] = b[k]
            else:
                a[k].modify(v)
        else:
            if affix == Affix.PRE:
                a.data = {k: v, **a.data}
            elif affix == Affix.POST:
                a.data = {**a.data, k: v}
            elif prev_key is None:
                a.data = {**a.data, k: v}
            else:
                a.data = _insert(a, prev_key, k, v)
        prev_key = k
    return a.data


def merge(*args: Tuple[dict]) -> ConfigDict:
    """Merges different configs

    Returns:
        ConfigDict: merged configs
    """

    return reduce(_modify, map(_config_factory, args), ConfigDict({}))



from dicfg.addons.addon import UpdaterAddon
from dicfg.config import Affix, ConfigDict, ConfigList, ConfigValue, update


class ReplaceUpdaterAddon(UpdaterAddon):

    NAME = "replace"

    @classmethod
    def update(cls, a: ConfigValue, b: ConfigValue):
        if isinstance(a, ConfigDict):
            return b.data
        return update(a, b)


class MergeUpdaterAddon(UpdaterAddon):

    NAME = "merge"

    @classmethod
    def update(cls, a: ConfigValue, b: ConfigValue):
        if isinstance(a, ConfigList):
            return a.data + b.data
        return update(a, b)


class PreUpdaterAddon(UpdaterAddon):

    NAME = "pre"

    @classmethod
    def update(cls, a: ConfigValue, b: ConfigValue):
        return update(a, b, affix=Affix.PRE)


class PostUpdaterAddon(UpdaterAddon):

    NAME = "post"

    @classmethod
    def update(cls, a: ConfigValue, b: ConfigValue):
        return update(a, b, affix=Affix.POST)

class NonEmptyUpdater(UpdaterAddon):
    
    NAME = "~"

    @classmethod
    def update(cls, a: ConfigValue, b: ConfigValue):
        return a



import operator
import os
import re
from copy import deepcopy
from functools import reduce, singledispatchmethod
from importlib import import_module
from typing import Union
from dicfg.addons.templates import LogLevel

REFERENCE_START_SYMBOL = "$"
REFERENCE_MAP_SYMBOL = ":"
REFERENCE_ATTRIBUTE_SYMBOL = "."

OBJECT_KEY = "*object"
ARGS_KEY = "*args"
KWARGS_KEY = "**kargs"
BUILD_KEY = "*build"
LOG_KEY = "*log"

WHITE_LIST_FACTORY_KEYS = [OBJECT_KEY, ARGS_KEY, KWARGS_KEY, BUILD_KEY, LOG_KEY]


class InvalidObjectConfigurationError(Exception):
    """Custom exception for invalid object configuration errors."""

    pass

class InvalidReferenceError(Exception):
    """Custom exception for invalid reference errors."""

    pass


class _ObjectFactory:
    def __init__(self, config: dict):
        self._configuration = config

        self._re_pattern_map = {
            "\\${\\$env.(.*)}": os.environ.get,
            "\\${\\$(.*)}": self._parse_object_str,
        }

    def build_config(self):
        return self._build(self._configuration)

    @singledispatchmethod
    def _build(self, config: dict):
        return config

    @_build.register
    def _build_dict(self, config: dict):
        for key, value in config.items():
            if not _build(value):
                config[key] = value
                continue

            if _is_object_config(value):
                self._logging(value, key)
                config[key] = self._build_object(value)
            else:
                config[key] = self._build(value)

        return config

    def _logging(self, value: dict, key: str):
        """Handle logging based on the configuration value."""
        if log := value.pop(LOG_KEY, False):
            if log == LogLevel.VERBOSE:
                print("\tRunning:", key)
            elif log == LogLevel.DEBUG:
                print("\tRunning:", key)
                print("\tConfig:", value)
            else:
                raise ValueError(f"Invalid log level: {log}")

    @_build.register(list)
    @_build.register(tuple)
    def _build_list(self, config: Union[list, tuple]):
        for idx, item in enumerate(config):
            if _is_object_config(item):
                config[idx] = self._build_object(item)
            else:
                config[idx] = self._build(item)
        return config

    @_build.register
    def _build_str(self, config: str):
        if config.lower() == "none":
            return None
        if REFERENCE_START_SYMBOL in config:
            return self._get_reference(reference=config)
        return config

    def _build_object(self, value: dict):
        try:
            kwargs = self._build(value)
            object_string = value.pop(OBJECT_KEY)
            args = value.pop(ARGS_KEY, ())
            kwargs.update(value.pop(KWARGS_KEY, {}))
            attribute = self._parse_object_str(object_string)
            return attribute(*args, **kwargs)
        except (ModuleNotFoundError, AttributeError, TypeError) as e:
            raise InvalidObjectConfigurationError(
                f"Invalid object configuration:\nObject: {object_string}\nArgs: {value}\nError: {e}"
            ) from e

    def _parse_object_str(self, object_string: str):
        object_split = object_string.split(".")
        module_string = ".".join(object_split[:-1])
        attribute_string = object_split[-1]
        module = import_module(module_string)
        return getattr(module, attribute_string)

    def _get_reference(self, reference: str):
        try:
            for pattern, parse in self._re_pattern_map.items():
                match = re.fullmatch(pattern, reference)
                if match is not None:
                    return parse(match.group(1))

            matches = re.findall("\\${(.*?)}", reference)
            if len(matches) == 1 and len(matches[0]) + 3 == len(reference):
                return self._object_interpolation(matches[0])
            return self._string_interpolation(reference, matches)
        except (IndexError, KeyError, AttributeError):
            raise InvalidReferenceError(reference)

    def _object_interpolation(self, reference: str):
        references = reference.split(REFERENCE_MAP_SYMBOL)
        reference = reduce(operator.getitem, references[:-1], self._configuration)
        attributes = references[-1].split(REFERENCE_ATTRIBUTE_SYMBOL)
        reference = reference[attributes[0]]
        for attr in attributes[1:]:
            reference = getattr(reference, attr)
        return reference

    def _string_interpolation(self, reference, matches):
        for match in matches:
            _reference = self._object_interpolation(match)
            match = "${" + match + "}"
            reference = reference.replace(match, str(_reference))
        return reference


def _is_object_config(value):
    return isinstance(value, dict) and OBJECT_KEY in value


def _build(value):
    if isinstance(value, dict):
        return value.pop(BUILD_KEY, True)
    return True


def build_config(config: dict):
    """Builds config

    Args:
        config (dict): config to build

    Returns:
        dict: build config
    """

    return _ObjectFactory(deepcopy(config)).build_config()



from abc import ABC, abstractmethod
from enum import Enum
import re


class CONFIG_ADDONS(Enum):
    UPDATER = "updater"
    VALIDATOR = "validator"
    TEMPLATE = "template"


# Constants for addon signs
VALIDATOR_SIGN = "!"
UPDATER_SIGN = "&"
TEMPLATE_SIGN = "#"

# Pattern to match addon annotations
_ADDON_PATTERN = rf"@(\w+)\(([^)]+)\)|([{VALIDATOR_SIGN}{UPDATER_SIGN}{TEMPLATE_SIGN}])(\w+)"


def process_addons(key: str):
    """
    Extracts addons and their arguments from the key.
    Removes all addon annotations from the key and returns the cleaned key.
    
    :param key: The input key string with addon annotations.
    :return: A tuple of cleaned key and a list of extracted addons.
    """
    matches = re.findall(_ADDON_PATTERN, key)
    addons = []
    
    for match in matches:
        if match[0]:  # Match for @addon(name)
            addons.append((match[0], match[1]))
        elif match[2] == VALIDATOR_SIGN:  
            addons.append((CONFIG_ADDONS.VALIDATOR.value, match[3]))
        elif match[2] == UPDATER_SIGN:  
            addons.append((CONFIG_ADDONS.UPDATER.value, match[3]))
        elif match[2] == TEMPLATE_SIGN:  
            addons.append((CONFIG_ADDONS.TEMPLATE.value, match[3]))

    # Remove all annotations from the key
    key = re.sub(_ADDON_PATTERN, "", key)
    # Strip any leading/trailing whitespace or separators (if needed)
    key = key.strip()
    return key, addons


class UnsupportedAddonError(Exception):
    """Exception raised when an unsupported addon is used"""


class Addon(ABC):
    """Base class for config validators"""

    __registry = None
    NAME = None

    def __init_subclass__(cls: "Addon", **kwargs):
        if cls.__registry is None:
            cls.__registry = {}

        if cls.NAME in cls.__registry:
            raise ValueError(f"Addon with name '{cls.NAME}' already exists.")

        cls.__registry[cls.NAME] = cls
        super().__init_subclass__(**kwargs)

    @classmethod
    def get_addon(cls, name: str) -> "Addon":
        if name not in cls.__registry:
            raise UnsupportedAddonError(
                f"Addon {cls.NAME} with name '{name}' not found."
            )
        return cls.__registry[name]


class TemplateAddon(Addon):
    """Template addon for config values to be used for predefined templates"""

    NAME = "template"

    @classmethod
    def data(cls):
        dat = {"*object": cls._get_object_ref()}
        dat.update(cls._data())
        return dat
    
    @classmethod
    def _get_object_ref(cls):
        return f"{cls.__module__}.{cls.__name__}"

    @classmethod
    def _data(cls):
        """Data to be used for template"""


class ValidatorAddon(Addon):
    """Validator addon for config values to be used to validate data"""

    NAME = "validator"

    @classmethod
    def validate(cls, value):
        """Validate the value"""


class UpdaterAddon(Addon):
    """Replace addon for config values to be used to replace or merge data"""

    NAME = "updater"

    @classmethod
    def update(cls, a, b):
        """Update a with b"""


_ADDONS = {
    CONFIG_ADDONS.UPDATER: UpdaterAddon,
    CONFIG_ADDONS.VALIDATOR: ValidatorAddon,
    CONFIG_ADDONS.TEMPLATE: TemplateAddon,
}


def select_addon(addon: str, name: str) -> Addon:
    try:
        _addon: Addon = _ADDONS[CONFIG_ADDONS(addon)]
    except ValueError:
        raise UnsupportedAddonError(f"Addon {addon} not found.")
    return _addon.get_addon(name)



import ast
import sys
from collections import defaultdict
from copy import deepcopy
from functools import partial, singledispatch
from pathlib import Path
from typing import List, Optional, Union
from dicfg.config import merge
from dicfg.addons.validators import ValidationErrors
from dicfg.addons import load as _
from dicfg.formats import FORMAT_READERS
from pprint import pprint

class ConfigNotFoundError(Exception):
    """Raised when config file can not be found."""


class ConfigReader:
    """ConfigReader

    Args:
        name (str): Name of config. Used as a reference in user configs and cli settings.
        main_config_path (Union[str, Path], optional): Path to main config. Defaults to  "./configs/config.yml".
        presets_folder_name (str, optional): Presets folder. Defaults to 'presets'.
        default_key (str, optional): Default context key. Defaults to "default".
        context_keys (tuple, optional): Addtional context keys. Defaults to ().
        search_paths (tuple, optional): Search paths for config file interpolation. Defaults to ().
    """

    def __init__(
        self,
        name: str,
        main_config_path: Union[str, Path] = "./configs/config.yml",
        presets: Union[str, Path] = "presets",
        default_key: str = "default",
        context_keys: tuple = (),
        search_paths: tuple = (),
    ):
        self._name = name
        self._main_config_path = Path(main_config_path)

        if not self._main_config_path.exists():
            raise ConfigNotFoundError(
                f"No main config file found at: {self._main_config_path}. The default main config path can be set via the 'main_config_path argument'"
            )

        self._default_key = default_key
        self._context_keys = context_keys
        self._search_paths = search_paths

        self._configs_folder = None
        self._presets_folder = None

        self._configs_folder = self._main_config_path.parent

        if isinstance(presets, Path):
            self._presets_folder = presets
        else:
            self._presets_folder = self._configs_folder / presets

    def read(
        self,
        user_config: Optional[Union[dict, str, Path, list[dict, str, Path]]] = None,
        presets: tuple = (),
    ) -> dict:
        """Reads Config File

        Args:
            user_config (Union[dict, str, Path], optional): user_config Defaults to None.
            presets (tuple, optional): presets Defaults to ().

        Returns:
            dict: read configs
        """

        self_config = self._read(self._main_config_path)
        arg_preset_configs = self._read_presets(presets)
        cli_config = self._read_cli(sys.argv[1:])

        user_configs = []
        user_presets_configs = []
        user_config_search_paths = []
        if user_config is not None:
            if not isinstance(user_config, (list, tuple)):
                user_config = [user_config]

            for config in user_config:
                if isinstance(config, (str, Path)) and not isinstance(config, dict):
                    user_config_search_path = Path(config).parent
                    user_config_search_paths.append(user_config_search_path)
                read_user_config = self._read_user_config(config)
                user_presets = read_user_config.pop("presets", ())
                user_configs.append(read_user_config)
                user_presets_configs.extend(self._read_presets(user_presets))

        configs = (
            self_config,
            *tuple(user_presets_configs),
            *tuple(arg_preset_configs),
            *tuple(user_configs),
            cli_config,
        )

        search_paths = self._set_search_paths(
            user_config_search_paths, self._search_paths
        )

        configs = self._fuse_configs(configs, self._context_keys, search_paths)
        pprint(configs)
        merged_configs = merge(*configs)
        
        if errors := list(merged_configs.validate()):
            raise ValidationErrors(errors)
        return merged_configs.cast()

    def _set_search_paths(self, user_config_search_paths, search_paths):
        return (
            Path(),
            *tuple(user_config_search_paths),
            self._configs_folder,
            self._presets_folder,
            *search_paths,
        )

    def _read(self, config_path):
        config = FORMAT_READERS[Path(config_path).suffix](config_path=config_path)
        return {} if config is None else config

    def _read_presets(self, presets):
        return tuple((self._read(self._presets_folder / preset) for preset in presets))

    def _read_user_config(self, user_config):
        user_config = (
            user_config if isinstance(user_config, dict) else self._read(user_config)
        )

        try:
            return user_config[self._name]
        except KeyError:
            raise KeyError(
                f"Config file: {user_config} does not contain a '{self._name}' key."
            )

    def _read_cli(self, args: List[str]):
        dicts = []
        for arg in args:
            if arg.startswith(self._name) and "=" in arg:
                keys, value = arg.split("=")
                keys = keys.split(".")
                dicts.append(_create_dict_from_keys(keys, value))
        cli_config = merge(*dicts)
        return cli_config.get(self._name, {})

    def _fuse_configs(self, configs, context_keys, search_paths):
        fuse_config = partial(
            self._fuse_config, context_keys=context_keys, search_paths=search_paths
        )
        return tuple(map(fuse_config, configs))

    def _fuse_config(self, config: dict, context_keys: tuple, search_paths):
        config = _include_configs(config, search_paths)
        fused_config = deepcopy(
            {key: deepcopy(config.get("default", {})) for key in context_keys}
        )
        return merge(fused_config, config)


def _create_dict_from_keys(keys: list, value) -> dict:
    dictionary = defaultdict(dict)
    if len(keys) <= 1:
        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            value = ast.literal_eval("'" + value + "'")
        dictionary[keys[0]] = value
    else:
        dictionary[keys[0]] = dict(_create_dict_from_keys(keys[1:], value))
    return dict(dictionary)


def _search_config(config_name: Union[str, Path], search_paths: tuple) -> Path:
    for search_path in search_paths:
        if search_path is None:
            continue
        config_path = Path(search_path) / config_name
        if config_path.exists():
            return config_path
    raise ConfigNotFoundError(config_name)


@singledispatch
def _include_configs(config, search_paths):
    return config


@_include_configs.register
def _include_configs_str(config: str, search_paths):
    if Path(config).suffix in FORMAT_READERS:
        config_path = _search_config(config, search_paths)
        open_config = FORMAT_READERS[Path(config_path).suffix](config_path)
        return _include_configs(open_config, search_paths)
    return config


@_include_configs.register
def _include_configs_dict(config: dict, search_paths):
    for key, value in config.items():
        config[key] = _include_configs(value, search_paths)
    return config



from dataclasses import dataclass
import importlib

import inspect

from dicfg.addons.addon import ValidatorAddon
from dicfg.factory import WHITE_LIST_FACTORY_KEYS, OBJECT_KEY


@dataclass(frozen=True)
class ValidationError:
    message: str

class ValidationErrors(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return "\n".join(str(err) for err in self.errors)


class DependenciesAddon(ValidatorAddon):
    NAME = "dependencies"
    

class RequiredValidatorAddon(ValidatorAddon):
    """Validator that checks if a value is not empty or None"""

    NAME = "required"

    @classmethod
    def validate(cls, value):
        if not (value != "" and value is not None):
            return ValidationError("Value is required and must not be empty or None.")

class DepreciatedValidatorAddon(ValidatorAddon):
    """Validator that checks if a value is not empty or None"""

    NAME = "depreciated"

    @classmethod
    def validate(cls, value):
        if not (value == "" or value is None):
            return ValidationError("Value is depreciated and should not be used")


class ObjectValidatorAddon(ValidatorAddon):
    """Validator that checks if a value is a valid object configuration"""

    NAME = "object"

    @classmethod
    def validate(cls, value: dict) -> ValidationError:
        if not isinstance(value, dict):
            return ValidationError("Value must be a dictionary.")

        if OBJECT_KEY not in value:
            return ValidationError(
                f"The key {OBJECT_KEY} must be present in the configuration."
            )

        object_path = value[OBJECT_KEY].cast()
        try:
            module_name, object_name = object_path.rsplit(".", 1)
            module = importlib.import_module(module_name)
            obj = getattr(module, object_name)
            if not callable(obj):
                return ValidationError(f"'{object_path}' is not callable.")
        except (ImportError, AttributeError, ValueError) as e:
            return ValidationError(
                f"Failed to import or access callable for {OBJECT_KEY}: {str(e)}"
            )

        sig = inspect.signature(obj)

        remaining_keys = {
            key: value[key] for key in value if key not in WHITE_LIST_FACTORY_KEYS
        }

        for key in remaining_keys:
            if key not in sig.parameters:
                return ValidationError(
                    f"'{key}' is not a valid argument for '{object_path}'."
                )


import json
import yaml


def read_json_config(config_path):
    with open(str(config_path), encoding="utf8") as file:
        return json.load(file)


def read_yaml_config(config_path):
    with open(str(config_path), encoding="utf8") as file:
        return yaml.load(file, Loader=yaml.SafeLoader)


FORMAT_READERS = {
    ".json": read_json_config,
    ".yml": read_yaml_config,
    ".yaml": read_yaml_config,
}



from dicfg.addons.addon import TemplateAddon
from enum import Enum

class LogLevel(Enum):
    """Log level for templates"""
    DEBUG = "DEBUG"
    VERBOSE = "VERBOSE"


class Verbose(TemplateAddon):
    """Template"""

    NAME = "verbose"

    @classmethod
    def data(self):
        return {
            '*log': LogLevel.VERBOSE
        }


class Debug(TemplateAddon):
    """Template"""

    NAME = "debug"

    @classmethod
    def data(self):
        return {
            '*log': LogLevel.DEBUG
        }

class Build(TemplateAddon):
    """Template"""

    NAME = "build"

    @classmethod
    def data(self):
        return {
            '*build': True
        }

class DontBuild(TemplateAddon):
    """Template for io.StringIO object"""

    NAME = "dontbuild"

    @classmethod
    def data(self):
        return {
            '*build': False
        }



import yaml
from IPython import get_ipython
from IPython.core.magic import Magics, cell_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring


@magics_class
class ConfigMagics(Magics):
    """
    IPython magic for loading YAML configuration files.
    """

    @classmethod
    def register_magics(cls):
        """
        Register the magics with IPython.
        """
        ip = get_ipython()
        ip.register_magics(cls)

    @cell_magic
    @magic_arguments()
    @argument("config", help="The name to give the loaded configuration object.")
    def yaml(self, args, content):
        """
        Load a YAML configuration file into a variable in the IPython namespace.
        """

        args = parse_argstring(self.yaml, args)
        try:
            config = yaml.safe_load(content)
        except yaml.YAMLError as exc:
            print(f"Error parsing YAML: {exc}")
            return
        self.shell.user_ns[args.config] = config
